/*
 * Copyright (C) 2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// startsync.cpp - main module for the sync application
//

// Windows Header Files:
#include <windows.h>
#include <initguid.h>
#include <Notify.h>

#include "base/startcmd.h"
#include "base/util/utils.h"

#include "pim/SettingFunctions.h"
#include "pim/maincpp.h"
#include "notify/addresschange.h"

#include "base/Log.h"

#include "resource.h"

#define DIMCONF 128

//--------------------------------------------------------------------- Prototypes
static inline Alert(const wchar_t *msg, const wchar_t *title=TEXT("Warning"));

static int getPollingTime();
static bool isServerNotified();
static void setServerNotified(bool val);

static int polling(const wchar_t **sources);
static int sync(const wchar_t **sources);

//--------------------------------------------------------------- Static Functions

static inline Alert(const wchar_t *msg, const wchar_t *title)
{
#ifdef DEBUG
  	MessageBox (NULL, msg, title, MB_SETFOREGROUND |MB_OK);
#endif
}

static int getPollingTime()
{
    wchar_t buf[32];
    
    wmemset(buf, 0, 32);
    getClientConfigurationInternal (TEXT(""), PROPERTY_SPDM_POLLING_NOTIFICATION, 
                                 buf, FALSE);
    return wcstol(buf, NULL, 10);
}

static bool isServerNotified()
{
    wchar_t buf[10];

    wmemset(buf, 0, 10);
    getClientConfigurationInternal (TEXT(""), PROPERTY_SPDM_SERVER_NOTIFIED, 
                                     buf, FALSE);
    return ( wcscmp(buf, TEXT("1")) == 0);
        
}

static void setServerNotified(bool val)
{
    setClientConfiguration(
            TEXT(""),
            PROPERTY_SPDM_SERVER_NOTIFIED,
            ( (val) ? TEXT("1") : TEXT("0") ),
            NULL);
}

/**
 * Start polling on the specified sources
 */
static int polling(const wchar_t **sources=NULL)
{    
    long t = getPollingTime();
    int ret = 0;

    if (t > 0 && !isServerNotified()) {
        LOG.debug(L"sync");
        ret = sync(sources);
        LOG.debug(L"sleep");
        Sleep(t*1000*60);
        LOG.debug(L"wakeup");
        t = getPollingTime();
        //
        // reschedule and exit.
        //
        if (t > 0) {
            wchar_t cmdLine[512];
            wmemset(cmdLine, 0, 512);
            wcscpy(cmdLine, TEXT("poll "));
            
            for (int i = 0; sources[i]; i++) {
                wcscat(cmdLine, sources[i]);
                wcscat(cmdLine, TEXT(","));
            }
            cmdLine[wcslen(cmdLine) -1] = NULL;

            startcmd(SYNCAPP, cmdLine);
        }        
    }    
    return ret;

}

/**
 * Syncronize the specified sources
 */
static int sync(const wchar_t **sources=NULL)
{
    HANDLE hMutex;//Handle to mutex...

    // Create mutex here
    hMutex = CreateMutex(NULL, TRUE, TEXT("FunSyncInProgress") );

    switch(GetLastError()) {
        case ERROR_SUCCESS:
            LOG.debug(L"Mutex created.");
            break;  
        case ERROR_ALREADY_EXISTS:    
            Alert(TEXT("Another sync in progress"), TEXT("Warning"));
            return -1;
        default:
        	Alert(TEXT("Failed to create mutex"), TEXT("Error"));
           return -1;
    }
    
    const wchar_t *path = getProgramPath();
    wchar_t errmsg[1024];

    //showNotification();
    int ret = synchronize(errmsg, path, (const wchar_t **)sources);
    //Sleep(2000);
    //hideNotification();

    CloseHandle( hMutex );
    LOG.debug(L"Mutex released.");

    return ret;
}

static int notify(const wchar_t **sources=NULL)
{
    AN_ResponseCode responseCode = notifyAddressChange(APPLICATION_URI);

    if(responseCode == AN_AddressAccepted ){
        setServerNotified(true);
    }
    else {
        setServerNotified(false);
        // Start the polling, if the polling time is set.
        return polling(sources);
    }
    return 0;
}

/**
 * Parse the command line
 *
 * @param cmdline - the command line (syntax: command [source,source,...])
 *
 * @param command (out) - the command name
 * @param sources (out) - a NULL terminated array of source names
 */
static int parseCmdLine(const LPTSTR cmdline,
                        wchar_t **command,
                        wchar_t ***sources)
{
    *command = stringdup(cmdline);
    *sources = NULL;

    wchar_t *args = wcschr(*command, ' ');
    if(!args) {
        // in this case comand == cmdline and sources is null
        return 0;
    }
    // terminate command and point to the start of arguments
    args[0] = 0; args++;
    if (!args[0]) {
        // empty args after a space, like before
        return 0;
    }
    int argc;
    wchar_t *arg = args;
    // Count arguments
    for(argc = 1; (arg=wcschr(arg+1, ',')) != 0; argc++);

    // Get arg list
    wchar_t **srclist = new wchar_t *[argc+1];
    arg=args;
    for(int i=0; i<argc; i++) {
        if(!arg || !arg[0])
            return -1;
        srclist[i]=arg;
 
        arg=wcschr(arg, ',');
        if(arg) {
            arg[0]=0; arg++;
        }
    }
    srclist[i]=0;
    *sources=srclist;

    return 0;
}


//-------------------------------------------------------------------------- MAIN

int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPTSTR    lpCmdLine,
					int       nCmdShow)
{
    wchar_t *command, **sources;
    int ret = 0;

    LOG.setLevel(LOG_LEVEL_DEBUG);

    //Alert(lpCmdLine);
    
    if (parseCmdLine(lpCmdLine, &command, &sources)) {
        Alert(TEXT("Syntax error in command line"), TEXT("StartSync"));
        return 1;
    }

    if(wcscmp(command, TEXT("poll")) == 0) 
        ret = polling((const wchar_t **)sources);
    else if (wcscmp(command, TEXT("once")) == 0)
        ret = sync((const wchar_t **)sources);
    else if (wcscmp(command, TEXT("addresschange")) == 0)
        ret = notify((const wchar_t **)sources);
    else
        Alert(TEXT("Unknown command."), TEXT("StartSync"));

    if (command)
        delete [] command;
    if (sources)
        delete [] sources;
    
	return ret;
}

